@@ -28,18 +28,19 @@ virtual circuit mode for large responses.
 
 package AnyEvent::DNS;
 
-no warnings;
-use strict;
-
+use Carp ();
 use Socket qw(AF_INET SOCK_DGRAM SOCK_STREAM);
 
-use AnyEvent ();
-use AnyEvent::Handle ();
+use AnyEvent (); BEGIN { AnyEvent::common_sense }
 use AnyEvent::Util qw(AF_INET6);
 
-our $VERSION = 4.41;
+our $VERSION = $AnyEvent::VERSION;
 
-our @DNS_FALLBACK = (v208.67.220.220, v208.67.222.222);
+# some public dns servers
+our @DNS_FALLBACK = (
+   v209.244.0.3, v209.244.0.4, # level3
+   v4.2.2.1, v4.2.2.3, v4.2.2.4, v4.2.2.5, v4.2.2.6, # vnsc-pri.sys.gtei.net
+);
 
 =item AnyEvent::DNS::a $domain, $cb->(@addrs)
 
@@ -297,7 +298,7 @@ EDNS0 in all requests.
 
 =cut
 
-our $EDNS0 = $ENV{PERL_ANYEVENT_EDNS0} * 1; # set to 1 to enable (partial) edns0
+our $EDNS0 = $ENV{PERL_ANYEVENT_EDNS0}*1; # set to 1 to enable (partial) edns0
 
 our %opcode_id = (
    query  => 0,
@@ -381,6 +382,13 @@ sub _enc_name($) {
    pack "(C/a*)*", (split /\./, shift), ""
 }
 
+if ($] < 5.008) {
+   # special slower 5.6 version
+   *_enc_name = sub ($) {
+      join "", map +(pack "C/a*", $_), (split /\./, shift), ""
+   };
+}
+
 sub _enc_qd() {
    (_enc_name $_->[0]) . pack "nn",
      ($_->[1] > 0 ? $_->[1] : $type_id {$_->[1]}),
@@ -506,7 +514,7 @@ our %dec_rr = (
     13 => sub { unpack "C/a* C/a*", $_ }, # hinfo
     15 => sub { local $ofs = $ofs + 2 - length; ((unpack "n", $_), _dec_name) }, # mx
     16 => sub { unpack "(C/a*)*", $_ }, # txt
-    28 => sub { AnyEvent::Socket::format_address ($_) }, # aaaa
+    28 => sub { AnyEvent::Socket::format_ipv6 ($_) }, # aaaa
     33 => sub { local $ofs = $ofs + 6 - length; ((unpack "nnn", $_), _dec_name) }, # srv
     35 => sub { # naptr
        # requires perl 5.10, sorry
@@ -651,14 +659,31 @@ calls.
 Unless you have special needs, prefer this function over creating your own
 resolver object.
 
+The resolver is created with the following parameters:
+
+   untaint          enabled
+   max_outstanding  $ENV{PERL_ANYEVENT_MAX_OUTSTANDING_DNS}
+
+C<os_config> will be used for OS-specific configuration, unless
+C<$ENV{PERL_ANYEVENT_RESOLV_CONF}> is specified, in which case that file
+gets parsed.
+
 =cut
 
 our $RESOLVER;
 
 sub resolver() {
    $RESOLVER || do {
-      $RESOLVER = new AnyEvent::DNS;
-      $RESOLVER->os_config;
+      $RESOLVER = new AnyEvent::DNS
+         untaint         => 1,
+         exists $ENV{PERL_ANYEVENT_MAX_OUTSTANDING_DNS}
+            ? (max_outstanding => $ENV{PERL_ANYEVENT_MAX_OUTSTANDING_DNS}*1 || 1) : (),
+      ;
+
+      exists $ENV{PERL_ANYEVENT_RESOLV_CONF}
+         ? length $ENV{PERL_ANYEVENT_RESOLV_CONF} && $RESOLVER->_parse_resolv_conf_file ($ENV{PERL_ANYEVENT_RESOLV_CONF})
+         : $RESOLVER->os_config;
+
       $RESOLVER
    }
 }
@@ -706,6 +731,11 @@ The number of seconds (default: C<300>) that a query id cannot be re-used
 after a timeout. If there was no time-out then query ids can be reused
 immediately.
 
+=item untaint => $boolean
+
+When true, then the resolver will automatically untaint results, and might
+also ignore certain environment variables.
+
 =back
 
 =cut
@@ -738,11 +768,11 @@ sub new {
 
       AnyEvent::Util::fh_nonblocking $fh4, 1;
       $self->{fh4} = $fh4;
-      $self->{rw4} = AnyEvent->io (fh => $fh4, poll => "r", cb => sub {
+      $self->{rw4} = AE::io $fh4, 0, sub {
          if (my $peer = recv $fh4, my $pkt, MAX_PKT, 0) {
             $wself->_recv ($pkt, $peer);
          }
-      });
+      };
    }
 
    if (AF_INET6 && socket my $fh6, AF_INET6, &Socket::SOCK_DGRAM, 0) {
@@ -750,11 +780,11 @@ sub new {
 
       $self->{fh6} = $fh6;
       AnyEvent::Util::fh_nonblocking $fh6, 1;
-      $self->{rw6} = AnyEvent->io (fh => $fh6, poll => "r", cb => sub {
+      $self->{rw6} = AE::io $fh6, 0, sub {
          if (my $peer = recv $fh6, my $pkt, MAX_PKT, 0) {
             $wself->_recv ($pkt, $peer);
          }
-      });
+      };
    }
 
    $got_socket
@@ -765,7 +795,7 @@ sub new {
    $self
 }
 
-=item $resolver->parse_resolv_conv ($string)
+=item $resolver->parse_resolv_conf ($string)
 
 Parses the given string as if it were a F<resolv.conf> file. The following
 directives are supported (but not necessarily implemented).
@@ -786,9 +816,9 @@ sub parse_resolv_conf {
    my $attempts;
 
    for (split /\n/, $resolvconf) {
-      if (/^\s*#/) {
-         # comment
-      } elsif (/^\s*nameserver\s+(\S+)\s*$/i) {
+      s/#.*$//; # not quite legal, but many people insist
+
+      if (/^\s*nameserver\s+(\S+)\s*$/i) {
          my $ip = $1;
          if (my $ipn = AnyEvent::Socket::parse_address ($ip)) {
             push @{ $self->{server} }, $ipn;
@@ -822,10 +852,21 @@ sub parse_resolv_conf {
    $self->_compile;
 }
 
+sub _parse_resolv_conf_file {
+   my ($self, $resolv_conf) = @_;
+
+   open my $fh, "<", $resolv_conf
+      or Carp::croak "$resolv_conf: $!";
+
+   local $/;
+   $self->parse_resolv_conf (<$fh>);
+}
+
 =item $resolver->os_config
 
-Tries so load and parse F</etc/resolv.conf> on portable operating systems. Tries various
-egregious hacks on windows to force the DNS servers and searchlist out of the system.
+Tries so load and parse F</etc/resolv.conf> on portable operating
+systems. Tries various egregious hacks on windows to force the DNS servers
+and searchlist out of the system.
 
 =cut
 
@@ -835,7 +876,7 @@ sub os_config {
    $self->{server} = [];
    $self->{search} = [];
 
-   if (AnyEvent::WIN32 || $^O =~ /cygwin/i) {
+   if ((AnyEvent::WIN32 || $^O =~ /cygwin/i)) {
       no strict 'refs';
 
       # there are many options to find the current nameservers etc. on windows
@@ -879,12 +920,10 @@ sub os_config {
          $self->_compile;
       }
    } else {
-      # try resolv.conf everywhere
+      # try resolv.conf everywhere else
 
-      if (open my $fh, "</etc/resolv.conf") {
-         local $/;
-         $self->parse_resolv_conf (<$fh>);
-      }
+      $self->_parse_resolv_conf_file ("/etc/resolv.conf")
+         if -e "/etc/resolv.conf";
    }
 }
 
@@ -941,6 +980,9 @@ sub _compile {
 sub _feed {
    my ($self, $res) = @_;
 
+   ($res) = $res =~ /^(.*)$/s
+      if AnyEvent::TAINT && $self->{untaint};
+
    $res = dns_unpack $res
       or return;
 
@@ -998,7 +1040,7 @@ sub _exec {
 
       my ($server, $timeout) = @$retry_cfg;
       
-      $self->{id}{$req->[2]} = [AnyEvent->timer (after => $timeout, cb => sub {
+      $self->{id}{$req->[2]} = [(AE::timer $timeout, 0, sub {
          $NOW = time;
 
          # timeout, try next
@@ -1014,6 +1056,8 @@ sub _exec {
                my ($fh) = @_
                   or return &$do_retry;
 
+               require AnyEvent::Handle;
+
                my $handle; $handle = new AnyEvent::Handle
                   fh       => $fh,
                   timeout  => $timeout,
@@ -1043,7 +1087,7 @@ sub _exec {
       
       my $sa = AnyEvent::Socket::pack_sockaddr (DOMAIN_PORT, $server);
 
-      my $fh = AF_INET == Socket::sockaddr_family ($sa)
+      my $fh = AF_INET == AnyEvent::Socket::sockaddr_family ($sa)
                ? $self->{fh4} : $self->{fh6}
          or return &$do_retry;
 
@@ -1068,10 +1112,10 @@ sub _scheduler {
 
       if (@{ $self->{reuse_q} } >= 30000) {
          # we ran out of ID's, wait a bit
-         $self->{reuse_to} ||= AnyEvent->timer (after => $self->{reuse_q}[0][0] - $NOW, cb => sub {
+         $self->{reuse_to} ||= AE::timer $self->{reuse_q}[0][0] - $NOW, 0, sub {
             delete $self->{reuse_to};
             $self->_scheduler;
-         });
+         };
          last;
       }
 
@@ -1276,7 +1320,7 @@ sub resolve($%) {
                      or return $do_search->(); # cname chain too long
 
                   $cname = 1;
-                  $name = $rr[0][3];
+                  $name = lc $rr[0][3];
 
                } elsif ($cname) {
                   # follow the cname
